# -*- coding: utf-8 -*-
"""
| ----------------------------------------------------------------------------------------------------------------
| Date                : August 2018
| Copyright           : (C) 2018 by Tinne Cahy (Geo Solutions) and Ann Crabbé (KU Leuven)
| Email               : tinne.cahy@geosolutions.be, ann.crabbe@kuleuven.be
| Acknowledgements    : Translated from Local Maximum Filter [C++ software]
|                       Ghent University, Laboratory of Forest Management and Spatial Information Techniques
|                       Lieven P.C. Verbeke
|
| This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
| License as published by the Free Software Foundation; either version 3 of the License, or any later version.
| ----------------------------------------------------------------------------------------------------------------
"""

import argparse
import os
import shutil
import osgeo.ogr as ogr
from localmaxfilter.scripts.local_max_filter import LocalMaxFilter
from localmaxfilter.in_out.imports import import_image, import_vector_as_image, import_vector_with_buffer, \
    import_intersection, import_extent_of_vector, import_image_by_mask_extent, check_path
from localmaxfilter.in_out.exports import write_point_layer_from_image_dict, write_as_shapefile,\
    write_voronoi_layer_from_pointlayer
from osgeo import osr
import osgeo.gdal as gdal
import tempfile

# Include code to work with processing algorithms
import sys
from qgis.core import QgsApplication
from qgis.analysis import QgsNativeAlgorithms

if not os.environ.get('QGIS_PREFIX_PATH'):
    if sys.platform.startswith('linux') or sys.platform.startswith('darwin'):
        os.environ['QGIS_PREFIX_PATH'] = '/usr'
    else:
        os.environ['QGIS_PREFIX_PATH'] = os.path.join("C:", os.sep, "OSGeo4W64", "apps", "qgis")
        # or "C:\\Program Files\\QGIS 3.4\\apps\\qgis"

QgsApplication.setPrefixPath(os.environ['QGIS_PREFIX_PATH'], True)
qgs = QgsApplication([], False)
qgs.initQgis()

# Append the path where processing plugin can be found
# Ubuntu Linux for example the prefix path is: /usr  and plugins: (qgis_prefix)/share/qgis/python/plugins
sys.path.append(os.path.join(os.environ['QGIS_PREFIX_PATH'], 'share', 'qgis', 'python', 'plugins'))  # unix, mac
# On Windows for example the prefix path is: C:\\OSGeo4W\\apps\\qgis\\ and plugins: (qgis_prefix)\\python\\plugins
sys.path.append(os.path.join(os.environ['QGIS_PREFIX_PATH'], 'python', 'plugins'))  # windows

import processing
from processing.core.Processing import Processing

Processing.initialize()
QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())


def create_parser():

    def _is_shapefile(fname):
        fname = str(fname)
        if not fname.endswith('.shp'):
            parser.error("The output filename is not a Shapefile (must end with '.shp')")
        return fname

    parser = argparse.ArgumentParser(description=LocalMaxFilter.__doc__)

    parser.add_argument('image', type=str, metavar='\b', help='path of the input reflectance image')

    parser.add_argument('sliding_window', type=float, metavar='\b',
                        help='length of the sliding window expressed in meters, '
                             'should be a multiple of the input image pixel size and be larger than one pixel')

    parser.add_argument('-m', '--mask', type=str, metavar='\b',
                        help='path of the overlapping mask image of the input reflectance image')

    parser.add_argument('-v', '--voronoi', action='store_true',
                        help='Create a Voronoi shapefile based on the tree location')

    parser.add_argument('-s', '--snaptool', type=float, metavar='\b',
                        help='Give distance [m] for the snaptool, should be max half of the window size. '
                             'Tree tops which are closer to each other than the given distance will be taken as one')

    parser.add_argument('-o', '--output', type=lambda s: _is_shapefile(s), metavar='\b',
                        help="output shapefile (default: in same folder with extension '_window_{}.shp'")

    return parser


def run_localmax(parser):
    """
    Documentation: localmaxfilter -h
    """
    # parse files
    args = parser.parse_args()

    # get input image
    input_file, metadata, projection = import_image(args.image, reflectance=False)

    # add mask
    area_raster = None
    if args.mask:
        # Validate projection system input raster and layer
        srs = osr.SpatialReference(wkt=metadata['projection'])
        if not srs.GetAttrValue('unit').lower() in ['metre', 'meter', 'metres', 'meters']:
            sys.exit(
                'The input image must have a projected coordination system '
                'in metres in order to calculate the tree density')
        # create buffer layer of area_layer
        buffer_layer_path = import_vector_with_buffer(args.mask, args.sliding_window, projection)
        # clip input raster by extent
        buffer_extent = import_extent_of_vector(buffer_layer_path)
        input_file, metadata = import_image_by_mask_extent(args.image, buffer_extent)
        # import area of interest
        area_raster = import_vector_as_image(buffer_layer_path, metadata)
        shutil.rmtree(os.path.dirname(buffer_layer_path))

    # snap_distance
    snap_distance = None
    if args.snaptool:
        snap_distance = args.snaptool
        if snap_distance > args.sliding_window/2:
            raise Exception('Snap distance should be max half of the window size')

    # filter local max using certain window size
    pixel_size = metadata['geo_transform'][1]
    window = int((args.sliding_window / pixel_size) // 2 * 2 + 1)
    if int(args.sliding_window / pixel_size) // 2 * 2 + 1 == 1:
        window = 3
    lmf_result = LocalMaxFilter(window).execute(input_file, area_of_interest=area_raster,
                                                snap=snap_distance, pixel_size=metadata['geo_transform'])

    # Write results to point layer file
    path = args.output or os.path.splitext(args.image)[0] + '_window_{}.shp'.format(args.sliding_window)
    name = "{0}_window_{1}".format(os.path.basename(path), window)
    # remove output if exists
    if os.path.exists(path):
        os.remove(path)

    point_layer_path = path if args.mask is None else os.path.join(tempfile.mkdtemp(), name + '_temp.shp')
    # check if the path to the folder exist
    check_path(os.path.dirname(os.path.abspath(path)))
    write_point_layer_from_image_dict(point_layer_path, lmf_result, name, metadata)

    if args.mask:
        # intersect with area of interest layer
        intersect_point_layer = import_intersection(point_layer_path, args.mask, output=path)
        print_output_per_polygon(args.mask, intersect_point_layer, window)

        # count trees for each polygon
    else:
        # count trees
        print("\n#########################################")
        print("\n# The input window size resulted in a window of {0} by {0} pixels".format(window))
        print("\n# Number of trees counted: {} ".format(len(lmf_result)))
        print("\n#########################################\n")

    if args.voronoi:
        print_output_voronoi(point_layer_path, args.image, args.mask, args.output)

        _remove_dir(point_layer_path)


def print_output_per_polygon(polygon_path, point_path, window):

    polygon_path_shp = polygon_path
    if not polygon_path.endswith('shp'):
        polygon_path_shp = os.path.join(
            tempfile.mkdtemp(), os.path.splitext(os.path.basename(polygon_path))[0] + '_temp.shp')
        write_as_shapefile(polygon_path_shp, polygon_path)

    polygon_data = ogr.Open(polygon_path_shp)
    polygon_layer = polygon_data.GetLayer()
    point_data = ogr.Open(point_path)
    point_layer = point_data.GetLayer()

    point_count = point_layer.GetFeatureCount()
    point_polygon_count = {}
    missing_id_message = None
    for i in range(0, point_count):
        if point_layer.GetFeature(i).GetField('id') is None:
            missing_id_message = "The input Shapefile misses some ID's, might result in misnamed ID's in the output"

    for i in range(0, point_count):
        p_id = point_layer.GetFeature(i).GetFieldAsInteger('id')
        point_polygon_count[p_id] = point_polygon_count.get(p_id, 0) + 1

    polygon_count = polygon_layer.GetFeatureCount()

    # print area, trees, density
    print("\n#########################################")
    print("\n# The input window size resulted in a window of {0} by {0} pixels".format(window))
    if missing_id_message is not None:
        print("\n# " + missing_id_message)
    for i in range(0, polygon_count):
        polygon_feature = polygon_layer.GetFeature(i)
        area = polygon_feature.GetGeometryRef().GetArea() / 10000
        id_polygon = polygon_feature.GetFieldAsInteger('id')
        count = point_polygon_count[id_polygon]
        print("\n# Area of interest with id {}".format(id_polygon))
        print("#   Area (ha): {} ".format(round(area, 4)))
        print("#   Number of trees counted: {} ".format(count))
        print("#   Tree density (trees/ha): {} ".format(round(count/area, 1)))
    print("\n#########################################\n")

    _remove_dir(polygon_path_shp)


def print_output_voronoi(point_layer, image_path, mask_layer=None, output=None):
    if mask_layer:
        voronoi_layer_path = write_voronoi_layer_from_pointlayer(point_layer)
        if output:
            voronoi_intersection_path = os.path.splitext(output)[0]+'_voronoi.shp'
            intersect_voronoi_path = import_intersection(voronoi_layer_path, mask_layer, voronoi_intersection_path)
        else:
            voronoi_intersection_path = os.path.splitext(image_path)[0]+'_voronoi.shp'
            intersect_voronoi_path = import_intersection(voronoi_layer_path, mask_layer, voronoi_intersection_path)
        print("\n#########################################")
        print("\n# Location of the Voronoi shapefile:")
        print(intersect_voronoi_path)
        print("\n#########################################\n")

    else:
        if output:
            voronoi_layer_path = os.path.splitext(output)[0]+'_voronoi.shp'
            write_voronoi_layer_from_pointlayer(point_layer, voronoi_layer_path)
        else:
            voronoi_layer_path = os.path.splitext(image_path)[0]+'_voronoi.shp'
            write_voronoi_layer_from_pointlayer(point_layer, voronoi_layer_path)
        print("\n#########################################")
        print("\n# Location of the Voronoi shapefile")
        print(voronoi_layer_path)
        print("\n#########################################\n")


def _remove_dir(path):
    try:
        shutil.rmtree(os.path.dirname(path))
    except:
        pass


def _unlink(path):
    try:
        gdal.Unlink(path)
    except:
        pass


def main():
    parser = create_parser()
    run_localmax(parser)


if __name__ == '__main__':
    main()
