# -*- coding: utf-8 -*-
"""
| ----------------------------------------------------------------------------------------------------------------
| Date                : August 2018
| Copyright           : (C) 2018 by Tinne Cahy (Geo Solutions) and Ann Crabbé (KU Leuven)
| Email               : tinne.cahy@geosolutions.be, ann.crabbe@kuleuven.be
| Acknowledgements    : Translated from Local Maximum Filter [C++ software]
|                       Ghent University, Laboratory of Forest Management and Spatial Information Techniques
|                       Lieven P.C. Verbeke
|
| This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
| License as published by the Free Software Foundation; either version 3 of the License, or any later version.
| ----------------------------------------------------------------------------------------------------------------
"""
import os
import sys
import shutil
from osgeo import osr
import osgeo.gdal as gdal
from qgis.PyQt.QtWidgets import QDialog, QFileDialog, QDialogButtonBox
from qgis.gui import QgsFileWidget
from qgis.core import QgsProviderRegistry, QgsMapLayerProxyModel, QgsRasterLayer, QgsProject, QgsVectorLayer
from qgis.PyQt.uic import loadUi
from localmaxfilter.scripts.local_max_filter import LocalMaxFilter
from localmaxfilter.gui.logo_gui import LogoWidget
from localmaxfilter.in_out.imports import import_image, import_vector_as_image, import_intersection, \
    import_vector_with_buffer, import_extent_of_vector, import_image_by_mask_extent
from localmaxfilter.in_out.exports import write_point_layer_from_image_dict, write_as_shapefile,\
    add_density_to_attribute_table, write_voronoi_layer_from_pointlayer, EmittingStream
import tempfile
import uuid


class LocalMaxFilterWidget(QDialog):
    """ QDialog to interactively set up the Local Max Filter input and output. """

    def __init__(self):
        super(LocalMaxFilterWidget, self).__init__()
        loadUi(os.path.join(os.path.dirname(__file__), 'local_max_filter.ui'), self)
        sys.stdout = EmittingStream(self.tabWidget)
        sys.stderr = EmittingStream(self.tabWidget)

        # Logo
        self.logoLayout.addWidget(LogoWidget(parent=self.logoWidget))

        # parameters
        excluded_providers = [p for p in QgsProviderRegistry.instance().providerList() if p not in ['gdal']]
        self.imageComboBox.setExcludedProviders(excluded_providers)
        self.imageComboBox.setFilters(QgsMapLayerProxyModel.RasterLayer)
        self.imageAction.triggered.connect(self._image_browse)
        self.imageButton.setDefaultAction(self.imageAction)

        excluded_providers = [p for p in QgsProviderRegistry.instance().providerList() if p not in ['ogr']]
        self.maskComboBox.setExcludedProviders(excluded_providers)
        self.maskComboBox.setFilters(QgsMapLayerProxyModel.VectorLayer)
        self.maskAction.triggered.connect(self._vector_browse)
        self.maskButton.setDefaultAction(self.maskAction)

        # set snap_distance maximum to half of the window size
        self.SNAPSpinBox.setMaximum(self.windowSpinBox.value()/2)
        self.windowSpinBox.valueChanged.connect(self._spinbox_value_changed)

        self.outputFileWidget.lineEdit().setReadOnly(True)
        self.outputFileWidget.lineEdit().setPlaceholderText('[Create temporary layer]')
        self.outputFileWidget.setStorageMode(QgsFileWidget.SaveFile)

        # run or cancel
        self.OKClose.button(QDialogButtonBox.Ok).setText("Run")
        self.OKClose.accepted.connect(self._run_local_max_filter)
        self.OKClose.rejected.connect(self.close)

        # widget variables
        self.image = None
        self.mask = None

    def __del__(self):
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    @staticmethod
    def write(s):
        print(s, file=sys.stderr)

    def _image_browse(self):
        """ Browse for an image raster file. """
        path = QFileDialog.getOpenFileName(filter=QgsProviderRegistry.instance().fileRasterFilters())[0]

        if len(path) > 0:
            gdal.UseExceptions()
            try:
                layer = QgsRasterLayer(path, os.path.basename(path), 'gdal')
                assert layer.isValid()
            except Exception:
                self.write("'" + path + "' not recognized as a supported file format.")
                return

            QgsProject.instance().addMapLayer(layer, True)
            self.imageComboBox.setLayer(layer)

    def _vector_browse(self):
        """ Browse for a vector file. """
        path = QFileDialog.getOpenFileName(filter=QgsProviderRegistry.instance().fileVectorFilters())[0]

        if len(path) > 0:
            gdal.UseExceptions()
            try:
                layer = QgsVectorLayer(path, os.path.basename(path), 'ogr')
                assert layer.isValid()
            except Exception:
                self.write("'" + path + "' not recognized as a supported file format.")
                return

            QgsProject.instance().addMapLayer(layer, True)
            self.maskComboBox.setLayer(layer)

    def _update_process_bar(self, value):
        self.progressBar.setValue(value)

    def _spinbox_value_changed(self, value):
        self.SNAPSpinBox.setMaximum(value/2)

    @staticmethod
    def run_algorithm_local_max_filter(layer, window_meter, update_process_bar, log_function, area_layer=None,
                                       output=None,  snap_distance=None, voronoi=None, feedback=None, context=None):

        # check if the image is not none
        if layer is None:
            raise Exception("Choose an image file.")

        # check if a former file with the same name is opened in QGIS, this obstruct the program to overwrite the files
        if output:
            former_shape = ('{0}_window_{1}.shp'.format(os.path.basename(output), window_meter))
            for layers in QgsProject.instance().layerTreeRoot().children():
                if layers.name() == former_shape:
                    raise Exception("Files which need to be removed are opened in QGIS and can't be removed. "
                                    "Create a new folder for your output or close the files in qgis")

        try:
            raster, metadata, projection = import_image(layer.source())
            # Rasterize input mask layer with buffer value
        except Exception as e:
            raise Exception(e)

        area_raster = None
        if area_layer:
            # Validate projection system input raster and layer
            srs = osr.SpatialReference(wkt=metadata['projection'])
            if not srs.GetAttrValue('unit').lower() in ['metre', 'meter', 'metres', 'meters']:
                raise Exception('The input image must have a projected coordination system '
                                'in metres in order to calculate the tree density')
            # create buffer layer of area_layer
            try:
                buffer_layer_path = import_vector_with_buffer(area_layer.source(), window_meter, projection,
                                                              feedback=feedback, context=context)
            except Exception as e:
                raise Exception(e)

            # create input raster by extent
            try:
                buffer_extent = import_extent_of_vector(buffer_layer_path)
            except Exception as e:
                raise Exception(e)
            try:
                raster, metadata = import_image_by_mask_extent(layer.source(), buffer_extent,
                                                               feedback=feedback, context=context)
            except Exception as e:
                raise Exception(e)
            # import area of interest
            try:
                area_raster = import_vector_as_image(buffer_layer_path, metadata)
            except Exception as e:
                raise Exception(e)
            _remove_dir(os.path.dirname(buffer_layer_path))

        # Run LMF, first convert the window size from meters to number of pixels (odd number)
        pixel_size = metadata['geo_transform'][1]
        window = int((window_meter / pixel_size) // 2 * 2 + 1)
        if int(window_meter / pixel_size) // 2 * 2 + 1 == 1:
            window = 3
        # print in the log the size of the window in pixels if the input length
        # doesn't match with an odd number of pixels
        if (window_meter / round(pixel_size, 2)) % 2 == 0 or (window_meter / pixel_size) // 2 * 2 + 1 == 1:
            log_function('The window is built out of {0} by {0} pixels'.format(window))
        # Run LMF
        lmf_result = LocalMaxFilter(window).execute(raster, area_of_interest=area_raster, snap=snap_distance,
                                                    pixel_size=metadata['geo_transform'], p=update_process_bar)
        update_process_bar(100)

        # Write results to point layer file
        if output is not None and len(output) != 0:
            name = os.path.basename(output) + '_window_{}'.format(window)
            #  NOT NECESSARY BECAUSE WE OVERWRITE. DO NOT USE WITHOUT CONSIDERATION COULD RESULT IN FILE LOSS
            #  remove output if exists
            # if os.path.exists(output):
            #    _remove_dir(output) if os.path.isdir(output) else os.remove(output)

        else:
            name = '{}_window_{}'.format(layer.name(), window)
            output = '/vsimem/' + uuid.uuid1().urn
        point_layer_path = output if area_layer is None else os.path.join(tempfile.mkdtemp(), name + '_temp.shp')
        try:
            write_point_layer_from_image_dict(point_layer_path, lmf_result, name, metadata)
        except Exception as e:
            raise Exception(e)

        area_layer_path = None
        if area_layer:
            # intersect with area of interest layer
            try:
                intersect_point_layer = import_intersection(point_layer_path, area_layer.source(),
                                                            feedback=feedback, context=context)
            except Exception as e:
                raise Exception(e)
            # create output layer
            area_layer_name = os.path.basename(area_layer.source())
            area_layer_path = '/vsimem/' + os.path.splitext(area_layer_name)[0] + '_window_{}.shp'.format(window)
            _unlink(area_layer_path)
            try:
                write_as_shapefile(area_layer_path, area_layer.source().rsplit('|')[0])
            except Exception as e:
                raise Exception(e)
                # add density to input area of interest layer
            try:
                add_density_to_attribute_table(area_layer_path, intersect_point_layer)
            except Exception as e:
                raise Exception(e)
            try:
                write_as_shapefile(output, intersect_point_layer, name)
            except Exception as e:
                raise Exception(e)
            _remove_dir(os.path.dirname(intersect_point_layer))

        # write point layer to Voronoi polygon layer
        voronoi_layer = None
        if voronoi:
            if '/vsimem/' in point_layer_path:
                input_voronoi = os.path.join(tempfile.mkdtemp(), name + '_temp.shp')
                try:
                    write_as_shapefile(input_voronoi, point_layer_path)
                except Exception as e:
                    raise Exception(e)
            else:
                input_voronoi = point_layer_path

            try:
                output_voronoi = write_voronoi_layer_from_pointlayer(input_voronoi, output=None,
                                                                     feedback=feedback, context=context)
            except Exception as e:
                raise Exception(e)
            _unlink('/vsimem/voronoi.shp')
            voronoi_layer = '/vsimem/{}_voronoi.shp'.format(os.path.basename(output))

            if area_layer:
                try:
                    intersect_voronoi = import_intersection(output_voronoi, area_layer.source(),
                                                            feedback=feedback, context=context)
                except Exception as e:
                    raise Exception(e)
                try:
                    write_as_shapefile(voronoi_layer, intersect_voronoi)
                except Exception as e:
                    raise Exception(e)
                _remove_dir(os.path.dirname(intersect_voronoi))
            else:
                try:
                    write_as_shapefile(voronoi_layer, output_voronoi)
                except Exception as e:
                    raise Exception(e)
            _remove_dir(os.path.dirname(output_voronoi))
            if point_layer_path.endswith('_temp.shp'):
                _remove_dir(os.path.dirname(point_layer_path))
        return output, area_layer_path, name, voronoi_layer

    def _run_local_max_filter(self):
        """ Read all parameters and pass them on to the LocalMaxFilter class. """

        # Only temp file possible when result is opened in QGIS
        if not self.openCheckBox.isChecked() and len(self.outputFileWidget.filePath()) == 0:
            self.write("If you won't open the result in QGIS, you must select an output file.")
            return

        # Get parameters
        try:
            layer = self.imageComboBox.currentLayer()
            window = self.windowSpinBox.value()
            area_layer = self.maskComboBox.currentLayer()
            snap_distance = self.SNAPSpinBox.value()
            path = self.outputFileWidget.filePath()
            voronoi = self.VoronoiCheckBox.isChecked()

            if path:
                output_path = os.path.split(path)[0]
            else:
                output_path = None

            path_out, area_layer_path, name, voronoi_layer = self.run_algorithm_local_max_filter(
                layer, window, self._update_process_bar, self.write, area_layer, output_path, snap_distance, voronoi)

        except Exception as e:
            self.write(e)
            return

        # Open result in QGIS
        if self.openCheckBox.isChecked():
            if path is not None and len(path) > 0:
                folder_name = os.path.basename(path)
            else:
                folder_name = 'OUTPUT'

            if area_layer:
                new_layer_polygon = QgsVectorLayer(area_layer_path, folder_name + '_' + os.path.basename(area_layer_path))
                QgsProject.instance().addMapLayer(new_layer_polygon, True)
            new_layer_point = QgsVectorLayer(path_out, folder_name + '_' + name)
            QgsProject.instance().addMapLayer(new_layer_point, True)
            if self.VoronoiCheckBox.isChecked():
                new_layer_voronoi = QgsVectorLayer(voronoi_layer, '{}_Voronoi'.format(folder_name))
                QgsProject.instance().addMapLayer(new_layer_voronoi, True)


def _remove_dir(path):
    try:
        shutil.rmtree(os.path.dirname(path))
    except:
        pass


def _unlink(path):
    try:
        gdal.Unlink(path)
    except:
        pass


def _run():
    from qgis.core import QgsApplication
    app = QgsApplication([], True)
    app.initQgis()

    z = LocalMaxFilterWidget()
    z.show()

    app.exec_()


if __name__ == '__main__':
    _run()
