# -*- coding: utf-8 -*-
"""
| ----------------------------------------------------------------------------------------------------------------
| Date                : March 2019
| Copyright           : (C) 2018 by Tinne Cahy, Louis Put (Geo Solutions) and Ann Crabbé (KU Leuven)
| Email               : tinne.cahy@geosolutions.be, ann.crabbe@kuleuven.be
| Acknowledgements    : Translated from Local Maximum Filter [C++ software]
|                       Ghent University, Laboratory of Forest Management and Spatial Information Techniques
|                       Lieven P.C. Verbeke
|
| This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
| License as published by the Free Software Foundation; either version 3 of the License, or any later version.
| ----------------------------------------------------------------------------------------------------------------
"""
from osgeo import ogr, osr
from localmaxfilter.in_out import check_path
import tempfile
import os
from qgis.PyQt.QtWidgets import QWidget


def write_point_layer_from_image_dict(output_file, image_dict, description, metadata):
    """ Convert the input dictionary image to a vector layer [output_path] using the image metadata.

    :param output_file: the absolute path to the output vector file
    :param image_dict: a dictionary containing the raster values and pixel locations.
        Keys: RasterVal, Pixel with keys x and y
    :param description: description of the output layer
    :param metadata: the metadata of the raster image.
        A dictionary with keys: geo_transform, x_size, y_size and projection
    """
    # https://gis.stackexchange.com/questions/268395/converting-raster-tif-to-point-shapefile-using-python

    # max distance between points
    origin_x = metadata['geo_transform'][0]
    origin_y = metadata['geo_transform'][3]
    pixel_width = metadata['geo_transform'][1]
    pixel_height = metadata['geo_transform'][5]

    # wkbPoint
    driver = ogr.GetDriverByName('ESRI Shapefile')
    datasource = driver.CreateDataSource(output_file)
    # make sure layer does not exists anymore in memory (necessary when using code from gui)
    if datasource.GetLayerByName(description):
        datasource.DeleteLayer(description)
    layer = datasource.CreateLayer(
        description,
        srs=osr.SpatialReference(wkt=metadata['projection']),
        geom_type=ogr.wkbPoint
    )
    feature_defn = layer.GetLayerDefn()
    layer.CreateField(ogr.FieldDefn("RasterVal", ogr.OFTInteger))

    # array2dict
    point = ogr.Geometry(ogr.wkbPoint)
    for image_val in image_dict:
        coord_x = origin_x + pixel_width * image_val['Pixel']['x'] + pixel_width / 2
        coord_y = origin_y + pixel_height * image_val['Pixel']['y'] + pixel_height / 2
        point.AddPoint(coord_x, coord_y)
        out_feature = ogr.Feature(feature_defn)
        out_feature.SetGeometry(point)
        out_feature.SetField("RasterVal", int(image_val['RasterVal']))
        layer.CreateFeature(out_feature)
        out_feature.Destroy()
    driver = None
    del layer, datasource


def write_as_shapefile(output_file, vector_path, name=None):
    """ Copy input of vector layer to Shapefile

    :param output_file: the absolute path to the output Shapefile
    :param vector_path: the absolute path to the input vector file
    :param name: name of the output layer (Optional)
    :return:
    """

    check_path(vector_path)

    vector_data = ogr.Open(vector_path)
    vector_layer = vector_data.GetLayer()

    driver = ogr.GetDriverByName('ESRI Shapefile')
    output_data = driver.CreateDataSource(output_file)
    name = name if name else output_file

    output_layer = output_data.CopyLayer(vector_layer, name, ['OVERWRITE=YES'])

    driver = None
    output_data.Destroy()
    del vector_layer, vector_data, output_layer, output_data


def add_density_to_attribute_table(polygon_path, point_path):
    """Add extra columns in polygon layer with area field 'Area_ha', tree count 'TreeCount' and tree density 'TreeDens'

    :param polygon_path: the absolute path to the input polygon vector file
    :param point_path: the absolute path to the input point vector file
    """

    check_path(polygon_path)
    check_path(point_path)
    polygon_path_id = None

    # It is only possible te edit an ESRI Shapefile (not an Idrisi vector layer)
    driver = ogr.GetDriverByName('ESRI Shapefile')

    polygon_data = driver.Open(polygon_path, 1)  # 0 means read-only. 1 means writeable.
    polygon_layer = polygon_data.GetLayer()
    point_data = driver.Open(point_path)
    point_layer = point_data.GetLayer()

    point_count = point_layer.GetFeatureCount()
    point_polygon_count = {}
    for i in range(0, point_count):
        if point_layer.GetFeature(i).GetField('id') is None:
            polygon_path_id = None
            break
        else:
            polygon_path_id = 'id'

    if polygon_path_id:
        point_count = point_layer.GetFeatureCount()
        point_polygon_count = {}
        for i in range(0, point_count):
            p_id = point_layer.GetFeature(i).GetFieldAsInteger(polygon_path_id)
            point_polygon_count[p_id] = point_polygon_count.get(p_id, 0) + 1
    else:
        point_wkts = [point_feature.GetGeometryRef().ExportToWkt() for point_feature in point_layer]

    # Attribute name TreeDensity is too long, therefore TreeDens
    field_config = {
        "Area_ha": {'width': 32, 'precision': 4},
        "TreeCount": {'width': 32, 'precision': 0},
        "TreeDens": {'width': 32, 'precision': 2}
    }
    for field_name in field_config.keys():
        if polygon_layer.GetLayerDefn().GetFieldIndex(field_name) <= 0:
            field = ogr.FieldDefn(field_name, ogr.OFTReal)
            field.SetWidth(field_config[field_name]['width'])
            field.SetPrecision(field_config[field_name]['precision'])
            polygon_layer.CreateField(field)

    polygon_count = polygon_layer.GetFeatureCount()
    for i in range(0, polygon_count):
        polygon_feature = polygon_layer.GetFeature(i)
        geom = polygon_feature.GetGeometryRef()
        area = geom.GetArea() / 10000
        polygon_feature.SetField("Area_ha", area)

        count = 0
        # Select the intersecting features
        if polygon_path_id:
            count = point_polygon_count[polygon_feature.GetFieldAsInteger(polygon_path_id)]
        else:
            for point_wkt in point_wkts:
                if ogr.CreateGeometryFromWkt(point_wkt).Intersects(polygon_feature.GetGeometryRef()):
                    count += 1

        polygon_feature.SetField("TreeCount", count)
        polygon_feature.SetField("TreeDens", count/area)
        polygon_layer.SetFeature(polygon_feature)


def write_voronoi_layer_from_pointlayer(point_layer, output=None, feedback=None, context=None):
    """Convert the input point vector to a Voronoi polygon layer.

    :param point_layer: the absolute path to the input point vector file
    :param feedback: necessary for the processing tool
    :param context: necessary for the processing tool
    :param output: optional output for the voronoi shapefile
    :return: returns the path of the voronoi shapefile
    """

    check_path(point_layer)

    if not output:
        tempdir = tempfile.mkdtemp()
        output = os.path.join(tempdir, 'voronoi.shp')

    # use processing in QGIS session
    import processing
    processing.run(
        "qgis:voronoipolygons",
        {'BUFFER': 0, 'INPUT': point_layer, 'OUTPUT': output},
        feedback=feedback,
        context=context
    )
    return output


class EmittingStream():
    """ QObject to catch the terminal output and send it along with q pyqtSignal. """

    def __init__(self, tab_widget):

        self.tab_widget = tab_widget
        self.log_index = tab_widget.indexOf(tab_widget.findChild(QWidget, 'tab_log'))
        self.log_widget = tab_widget.findChild(QWidget, 'logBrowser')

    def write(self, text):
        self.log_widget.append(text)
        self.tab_widget.setCurrentIndex(self.log_index)
