# -*- coding: utf-8 -*-
"""
| ----------------------------------------------------------------------------------------------------------------
| Date                : March 2019
| Copyright           : (C) 2018 by Tinne Cahy, Louis Put (Geo Solutions) and Ann Crabbé (KU Leuven)
| Email               : tinne.cahy@geosolutions.be, ann.crabbe@kuleuven.be
| Acknowledgements    : Translated from Local Maximum Filter [C++ software]
|                       Ghent University, Laboratory of Forest Management and Spatial Information Techniques
|                       Lieven P.C. Verbeke
|
| This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
| License as published by the Free Software Foundation; either version 3 of the License, or any later version.
| ----------------------------------------------------------------------------------------------------------------
"""
import osgeo.gdal as gdal
import osgeo.ogr as ogr
import osgeo.osr as osr
from localmaxfilter.in_out import check_path
import os
import shutil
import tempfile


def import_image(path, reflectance=False):
    """ Browse for a spectral image and return it without bad bands.

    :param path: the absolute path to the image
    :param reflectance: return reflectance values instead of DN between 0 and 254
    :return: float32 numpy array [#good bands x #rows x #columns]
    """

    check_path(path)

    gdal.UseExceptions()
    try:
        data = gdal.Open(path)
        array = data.ReadAsArray()
        metadata = {'geo_transform': data.GetGeoTransform(), 'x_size': data.RasterXSize, 'y_size': data.RasterYSize,
                    'projection': data.GetProjection()}

        projection = None
        srs = osr.SpatialReference(wkt=data.GetProjection())
        if srs.IsProjected:
            projection = srs.GetAttrValue('projcs')
            if srs.GetAttrValue('projcs') is None:
                raise Exception('Raster has no projection')

        if reflectance:
            array = array / float(255)  # adjust to reflectance values

        return array, metadata, projection

    except Exception as e:
        raise Exception(str(e))


def import_vector_as_image(path, metadata):
    """ Browse for a vector file and return it as a raster.

    :param path: the absolute path to the vector file
    :param metadata: the metadata of the output raster image.
        A dictionary with keys: geo_transform, x_size, y_size and projection
    :return: ogr layer
    """

    check_path(path)
    try:
        # assumes the same projection as metadata projection
        os.remove(path.replace('.shp', '.prj'))
        data = ogr.Open(path)
        layer = data.GetLayer()
        driver = gdal.GetDriverByName('MEM')
        memory_raster = driver.Create('', metadata['x_size'], metadata['y_size'], 1, gdal.GDT_Float32)
        memory_raster.SetGeoTransform(metadata['geo_transform'])
        memory_raster.SetProjection(metadata['projection'])

        gdal.RasterizeLayer(memory_raster, [1], layer,  burn_values=[1])  # , options=['ALL_TOUCHED=TRUE']
        array = memory_raster.ReadAsArray()

        return array

    except Exception as e:
        raise Exception(str(e))


def import_vector_with_buffer(path, length_window, projection, feedback=None, context=None):
    """ Browse for a vector file and return it as a buffered vector.

    :param path: the absolute path to the vector file
    :param length_window: the buffer distance given in meter
    :param projection: the projection of the raster
    :param feedback: necessary for the processing tool
    :param context: necessary for the processing tool
    :return: ogr layer
    """
    if '|' in path:
        path = path.partition('|')[0]
    data = ogr.Open(path)
    layer = data.GetLayer()
    spatialref = layer.GetSpatialRef()
    if spatialref.IsProjected:
        if spatialref.GetAttrValue('projcs') is None:
            raise Exception('Shapefile has no projection')
        elif spatialref.GetAttrValue('projcs') != projection:
            raise Exception('Raster and Shapefile have different projection')

    tempdir = tempfile.mkdtemp()
    output = os.path.join(tempdir, 'buffer.shp')

    buffer_distance = length_window * 2
    # use processing in qgis session
    import processing
    processing.run(
        "native:buffer",
        {
            'DISSOLVE': False, 'DISTANCE': buffer_distance, 'END_CAP_STYLE': 0, 'INPUT': path, 'JOIN_STYLE': 0,
            'MITER_LIMIT': 2, 'OUTPUT': output, 'SEGMENTS': 5
        },
        feedback=feedback,
        context=context
    )

    return output


def import_intersection(input_path, overlay_path, output=None, feedback=None, context=None):
    """ Returns the intersection between the input_path and the overlay_path

    :param input_path: the absolute path to the input vector file
    :param overlay_path: the absolute path to the overlay vector file
    :param output: the absolute path to the output vector file. This parameter is optional.
        When missing a layer is created in the temp folder
    :param feedback: necessary for the processing tool
    :param context: necessary for the processing tool
    :return: The absolute path to the output vector file.
    """
    if not output:
        tempdir = tempfile.mkdtemp()
        output = os.path.join(tempdir, 'intersection.shp')
    if '|' in overlay_path:
        overlay_path = overlay_path.partition('|')[0]

    # use processing in qgis session
    import processing
    processing.run(
        "native:intersection",
        {'INPUT': input_path, 'INPUT_FIELDS': [], 'OUTPUT': output, 'OVERLAY': overlay_path, 'OVERLAY_FIELDS': []},
        feedback=feedback,
        context=context
    )

    return output


def import_extent_of_vector(path):
    """ Returns the extent of a vector layer

    :param path: the absolute path to the input vector file
    :return: comma separated string with extent coordinates
    """
    check_path(path)
    # Get a Layer's Extent
    data_source = ogr.Open(path, 0)
    layer = data_source.GetLayer()
    return ', '.join(['{}'.format(e) for e in layer.GetExtent()])


def import_image_by_mask_extent(input_path, extent, feedback=None, context=None):
    """ Clips the input image raster by the extent and returns a numpy array with metada

    :param input_path: the absolute path to the input image
    :param extent: comma separated string with extent coordinates
    :param feedback: necessary for the processing tool
    :param context: necessary for the processing tool
    :return: float32 numpy array [#good bands x #rows x #columns] with metadata
    """
    tempdir = tempfile.mkdtemp()
    output = os.path.join(tempdir, 'clipraster.tif')

    # use processing in qgis session
    import processing
    processing.run(
        "gdal:cliprasterbyextent",
        {'INPUT': input_path, 'NODATA': -1, 'OPTIONS': '', 'OUTPUT': output, 'PROJWIN': extent},
        feedback=feedback,
        context=context
    )

    try:
        array, metadata, projection = import_image(output)
    except Exception as e:
        raise Exception(e)
    shutil.rmtree(os.path.dirname(output))
    return array, metadata
